import pickle
import torch
from torch.utils.data import Dataset, DataLoader
import torch.nn.functional as F
from tqdm.auto import tqdm
from typing import Tuple
import math
import numpy as np
import time
from functools import reduce
from operator import mul
from typing import Sequence, Tuple



# -----------------------------------------------------------------------------
#                  D a t a s e t   f o r   s l i d i n g   p a t c h e s
# -----------------------------------------------------------------------------


# ------------------------------ new channel-aware SlidingWindowND ------------------------------


class SlidingWindowND(Dataset):
    """
    Slide over a (D,H,W,C) array with 3-D windows; channel is not strided.

    volume : np.ndarray           shape (D,H,W,C)
    window : Sequence[int]        (d, h, w)
    stride : Sequence[int]        (sd, sh, sw)

    __getitem__ returns a torch.FloatTensor of shape (C, D, H, W)
    """
    def __init__(self, volume: np.ndarray, *, window: Sequence[int], stride: Sequence[int]):
        if volume.ndim != 4:
            raise ValueError("Expected volume with shape (D,H,W,C).")
        self.volume = volume
        self.window = tuple(int(x) for x in window)
        self.stride = tuple(int(x) for x in stride)

        (D,H,W,C) = volume.shape
        (d,h,w) = self.window
        (sd,sh,sw) = self.stride

        self.grid_shape: Tuple[int,int,int] = (
            math.floor((D - d) / sd) + 1,
            math.floor((H - h) / sh) + 1,
            math.floor((W - w) / sw) + 1,
        )
        self._n_patches = reduce(mul, self.grid_shape)

    def __len__(self) -> int:
        return self._n_patches

    def __getitem__(self, idx: int) -> torch.Tensor:
        # unravel idx -> (gz, gy, gx)
        gz = idx % self.grid_shape[0]
        tmp = idx // self.grid_shape[0]
        gy = tmp % self.grid_shape[1]
        gx = tmp // self.grid_shape[1]

        z0 = gz * self.stride[0]
        y0 = gy * self.stride[1]
        x0 = gx * self.stride[2]

        z1 = z0 + self.window[0]
        y1 = y0 + self.window[1]
        x1 = x0 + self.window[2]

        patch_dhwc = self.volume[z0:z1, y0:y1, x0:x1, :]  # (d,h,w,c)
        # return (C,D,H,W)
        patch_cd_hw = np.moveaxis(patch_dhwc, -1, 0)       # (c,d,h,w)
        return torch.from_numpy(patch_cd_hw).float()





class TraverseDataset2d(Dataset):
    def __init__(self, img: np.ndarray, stride: int, win_size: int, v: bool = False):
        """
        Args:
            img (np.ndarray): RGB image of shape [H, W, 3]
            stride (int): Sliding window stride
            win_size (int): Size of the square window
            v (bool): Verbose flag
        """
        assert img.ndim == 3 and img.shape[2] == 3, "Input must be an RGB image with shape [H, W, 3]"
        
        self.img = img
        self.v = v
        self.stride = stride
        self.win_size = win_size

        if v:
            print(f"init TraverseDataset2d with image of shape {img.shape}, stride = {stride}, win_size = {win_size}")

        self.patches = self._generate_patches()

    def _generate_patches(self):
        img = self.img
        patches = []
        for y in range(0, img.shape[0] - self.win_size + 1, self.stride):
            for x in range(0, img.shape[1] - self.win_size + 1, self.stride):
                patch = img[
                    y:y + self.win_size,
                    x:x + self.win_size,
                    :
                ]
                patches.append(patch)

        self.sample_shape = np.array([
            (img.shape[0] - self.win_size) // self.stride + 1,
            (img.shape[1] - self.win_size) // self.stride + 1,
        ])
        if self.v:
            print(f"sample shape = {self.sample_shape}")
        return patches

    def _get_sample_shape(self):
        return self.sample_shape

    def __len__(self):
        return len(self.patches)

    def __getitem__(self, idx):
        patch = self.patches[idx]
        patch = torch.tensor(patch, dtype=torch.float32).permute(2, 0, 1)  # Convert to [C, H, W]
        return patch

class TraverseDataset3d(Dataset):
    def __init__(self, img, stride: int, win_size, verbose=False):
        """
        Traverse a 3D volume using a sliding window.

        Args:
            img (ndarray): 3D input image of shape (Z, Y, X).
            stride (int): Stride for the sliding window.
            win_size (int or tuple/list of 3 ints): Size of the window in (Z, Y, X).
            verbose (bool): Print debug info if True.
        """
        self.img = img
        self.stride = stride
        self.win_size = self._normalize_win_size(win_size)
        self.verbose = verbose

        if self.verbose:
            print(f"Initializing TraverseDataset3d with shape {img.shape}, "
                  f"stride={stride}, win_size={self.win_size}")

        self.patches, self.sample_shape = self._generate_patches()

        if self.verbose:
            print(f"Sample shape: {self.sample_shape}, Total patches: {len(self.patches)}")

    def _normalize_win_size(self, win_size):
        """Ensure win_size is a tuple of 3 ints."""
        if isinstance(win_size, int):
            return (win_size,) * 3
        if isinstance(win_size, (list, tuple)) and len(win_size) == 3:
            return tuple(win_size)
        raise ValueError("win_size must be an int or a list/tuple of 3 ints.")

    def _generate_patches(self):
        """Extract 3D patches from the input image."""
        wz, wy, wx = self.win_size
        sz = (self.img.shape[0] - wz) // self.stride + 1
        sy = (self.img.shape[1] - wy) // self.stride + 1
        sx = (self.img.shape[2] - wx) // self.stride + 1

        patches = [
            self.img[z:z+wz, y:y+wy, x:x+wx]
            for z in range(0, self.img.shape[0] - wz + 1, self.stride)
            for y in range(0, self.img.shape[1] - wy + 1, self.stride)
            for x in range(0, self.img.shape[2] - wx + 1, self.stride)
        ]

        return patches, np.array([sz, sy, sx])

    def __len__(self):
        return len(self.patches)

    def __getitem__(self, idx):
        """Return a patch as a 1xD x H x W torch tensor."""
        patch = torch.from_numpy(self.patches[idx]).float().unsqueeze(0)
        return patch

    def get_sample_shape(self):
        return self.sample_shape

import numpy as np
import torch
from torch.utils.data import Dataset

class TraverseDataset3d_overlap(Dataset):
    def __init__(self, img, win_size, overlap=0, verbose=False):
        """
        Args:
            img (np.ndarray): 3D input image, shape (Z, Y, X)
            win_size (int or tuple/list of 3 ints): size of the 3D patch
            overlap (int or tuple/list of 3 ints): overlap between patches
            verbose (bool): if True, prints info during initialization
        """
        self.img = img
        self.verbose = verbose

        # Normalize win_size and overlap to 3-element lists
        self.win_size = self._expand_to_3d(win_size)
        self.overlap = self._expand_to_3d(overlap)

        # Compute stride
        self.stride = [w - o for w, o in zip(self.win_size, self.overlap)]
        if any(s <= 0 for s in self.stride):
            raise ValueError("Overlap must be smaller than win_size in all dimensions.")

        if self.verbose:
            print(f"Init TraverseDataset3d: img shape={img.shape}, win_size={self.win_size}, overlap={self.overlap}, stride={self.stride}")

        self.patches, self.roi_nums = self._generate_patches()

    def _expand_to_3d(self, val):
        if isinstance(val, int):
            return [val] * 3
        elif isinstance(val, (list, tuple)) and len(val) == 3:
            return list(val)
        else:
            raise ValueError("win_size and overlap must be either int or list/tuple of 3 ints")

    def _generate_patches(self):
        patches = []
        Z, Y, X = self.img.shape
        wz, wy, wx = self.win_size
        sz, sy, sx = self.stride

        for z in range(0, Z - wz + 1, sz):
            for y in range(0, Y - wy + 1, sy):
                for x in range(0, X - wx + 1, sx):
                    patch = self.img[z:z + wz, y:y + wy, x:x + wx]
                    patches.append(patch)

        roi_nums = [(img_dim - win + stride) // stride
                        for img_dim, win, stride in zip(self.img.shape, self.win_size, self.stride)]
        
        if self.verbose:
            print(f"Generated {len(patches)} patches with sample shape {roi_nums}")

        return patches, roi_nums

    def __len__(self):
        return len(self.patches)

    def __getitem__(self, idx):
        patch = torch.from_numpy(self.patches[idx]).float().unsqueeze(0)
        return patch

    def get_roi_nums(self):
        return self.roi_nums

def extract_feats(img_vol,win_size,cnn,mlp,stride,batch_size):
    """
    img_vol: need to be precropped
    """

    draw_border_dataset = TraverseDataset3d(img_vol,stride=stride,win_size=win_size)  
    border_draw_loader = DataLoader(draw_border_dataset,batch_size,shuffle=False,drop_last=False)
    print(f"len of dataset is {len(draw_border_dataset)}")

    current = time.time()
    feats_lst = get_feature_list('cuda',cnn,mlp,border_draw_loader,save_path=None)
    out_shape = draw_border_dataset.sample_shape

    print(f"extracting feature from image consume {time.time()-current} seconds")
    return feats_lst,out_shape


def get_feature_list(device,encoder,test_loader,extract_layer_name=None,save_path=None,adaptive_pool=False)->np.ndarray:
    """
    encoder inference on a single input 2d-image
    
    input(numpy)--> test_dataset
    collect feats during inference
    return the feats as shape of N*n_dim

    """
    print(f"device is {device}")


    # a dict to store the activations
    activation = {}
    def getActivation(name):
        # the hook signature
        def hook(model, input, output):
            activation[name] = output.detach()
        return hook

    #register the forward hook at layer"layer_name"
    if extract_layer_name:
        hook1 = getattr(encoder,extract_layer_name).register_forward_hook(getActivation(extract_layer_name))

    feats_list = []
    for i, imgs in enumerate(tqdm(test_loader,desc="extracting features")):
        outs=encoder(imgs.to(device))
        if adaptive_pool:
            outs = F.adaptive_avg_pool3d(outs, output_size=(1, 1, 1))
        B = outs.shape[0]
        if extract_layer_name:
            feats_list.append( np.squeeze(activation[extract_layer_name].cpu().detach().numpy().reshape(B,-1)))
        else:
            feats_list.append( np.squeeze(outs.cpu().detach().numpy().reshape(B,-1)))
    #detach the hook
    if extract_layer_name:
        hook1.remove()

    feats_array = np.concatenate([ arr for arr in feats_list], axis=0)
    print(f"fests_arry shape {feats_array.shape}")

    if save_path :
        with open(save_path, 'wb') as file:
            pickle.dump(feats_array, file)
    
    return feats_array






def get_feature_map(device,encoder,img = None,extract_layer_name=None,loader=None, overlap_i =32 ,roi_nums=None)->np.ndarray:
    """
    encoder inference on a single input 2d-image
    
    input(numpy)--> test_dataset
    collect feats during inference
    return the feats as shape of N*n_dim

    """
    print(f"device is {device}")


    # a dict to store the activations
    activation = {}
    def getActivation(name):
        # the hook signature
        def hook(model, input, output):
            activation[name] = output.detach()
        return hook

    #register the forward hook at layer"layer_name"
    if extract_layer_name:
        hook1 = getattr(encoder,extract_layer_name).register_forward_hook(getActivation(extract_layer_name))

    if isinstance(img, np.ndarray) : 
        input = torch.from_numpy(img).float().unsqueeze(0).to(device)
        outs=encoder(input)
        B = outs.shape[0]
        if extract_layer_name:
            feats_map =  np.squeeze(activation[extract_layer_name].cpu().detach().numpy())
        else:
            feats_map =  np.squeeze(outs.cpu().detach().numpy())
        #detach the hook
        if extract_layer_name:
            hook1.remove()

        return feats_map 
    else:

        test_batch_input = next(iter(loader)) 
        test_out = encoder(test_batch_input.to(device))
        # B should be 1
        C,H,W = np.squeeze(test_out.cpu().detach().numpy()).shape
        input_H = 1024
        input_W = 1024
        margin =int( np.ceil( overlap_i* H / input_H)) # cut margin at both side of H and W

        final_chunk_shape = (1,int((H-2*margin)),int((W-2*margin)),C )
        chunk_H = final_chunk_shape[1]
        chunk_W = final_chunk_shape[2]
        final_feats_shape = (1,final_chunk_shape[1]*roi_nums[1],final_chunk_shape[2]*roi_nums[2],C)
        final_feats_map = np.zeros(shape=final_feats_shape)


        for i, imgs in enumerate(tqdm(loader, desc='extracting features')):
            # batch =1

            outs=encoder(imgs.to(device))
            if extract_layer_name:
                feats_map =np.moveaxis( np.squeeze(activation[extract_layer_name].cpu().detach().numpy()), 0,-1)
            else:
                feats_map = np.moveaxis( np.squeeze(outs.cpu().detach().numpy()), 0,-1)

            feats_map = np.expand_dims(feats_map,axis=0) #(1,H,W,C)
            # Handle cropping depending on margin
            if margin > 0:
                feats_map_cropped = feats_map[:, margin:-margin, margin:-margin, :]
            else:
                feats_map_cropped = feats_map

            y = i // roi_nums[2]
            x = i % roi_nums[2]

            index = (
                slice(None),
                slice(y * chunk_H, (y + 1) * chunk_H),
                slice(x * chunk_W, (x + 1) * chunk_W),
                slice(None)
            )
            final_feats_map[index] = feats_map_cropped

        if extract_layer_name:
            hook1.remove()


        return np.squeeze(final_feats_map)



import numpy as np
import matplotlib.pyplot as plt
from skimage.segmentation import slic, mark_boundaries
from skimage import graph
from skimage import color
from scipy.ndimage import zoom
from sklearn.decomposition import PCA

from skimage.graph import RAG
import math

from confettii.plot_helper import grid_plot_list_imgs
def rag_mean_feature(image, labels, connectivity=2, mode='similarity', sigma=0.001):
    # Initialize RAG
    rag = RAG(labels, connectivity=connectivity)

    h, w, c = image.shape  # C is number of feature channels
    for n in rag:
        rag.nodes[n].update({
            'labels': [n],
            'pixel count': 0,
            'total feature': np.zeros((c,), dtype=np.float64),
        })

    # Accumulate feature sums and counts
    for index in np.ndindex(labels.shape):
        current = labels[index]
        rag.nodes[current]['pixel count'] += 1
        rag.nodes[current]['total feature'] += image[index]

    # Compute mean feature vector per region
    for n in rag:
        rag.nodes[n]['mean feature'] = (
            rag.nodes[n]['total feature'] / rag.nodes[n]['pixel count']
        )

    # Compute weights based on feature vector distance
    for x, y, d in rag.edges(data=True):
        diff = rag.nodes[x]['mean feature'] - rag.nodes[y]['mean feature']
        diff_norm = np.linalg.norm(diff)  # Euclidean distance
        if mode == 'similarity':
            d['weight'] = math.e ** (-(diff_norm**2) / sigma)
        elif mode == 'distance':
            d['weight'] = diff_norm
        else:
            raise ValueError(f"The mode '{mode}' is not recognised")

    return rag

def segment_and_plot_from_feats(
    feats_map,
    image,
    label = None,
    rag = None,
    slic_compactness=0.3,
    rag_weight_sigma=0.01,
    n_segments=100,
    slic_iters=30,
    ncut_thresh=0.001
):
    """
    Segments image using a feature map and visualizes PCA features, superpixels, RAG, and normalized cuts.

    Parameters:
    - feats_map: np.ndarray of shape (H, W, C) — Feature map
    - image: np.ndarray — Original image (H, W[, 3])
    - slic_compactness: float — Compactness for SLIC superpixels
    - rag_weight_sigma: float — Sigma for RAG edge weights
    - n_segments: int — Desired number of superpixels
    - slic_iters: int — Max iterations for SLIC
    - ncut_thresh: float — Threshold for normalized cut
    """
    
    # Match feature map resolution to input image
    zoom_factors = tuple(raw / feat for raw, feat in zip(image.shape[:2], feats_map.shape[:2]))
    feats_map_rescaled = zoom(feats_map, zoom=(*zoom_factors, 1), order=1)

    # Normalize features
    feats_map_rescaled = (feats_map_rescaled - feats_map_rescaled.min()) / (feats_map_rescaled.max() - feats_map_rescaled.min())
    
    # SLIC Superpixels
    if label is None: 
        label = slic(
            feats_map_rescaled,
            n_segments=n_segments,
            compactness=slic_compactness,
            max_num_iter=slic_iters,
            start_label=1,
            channel_axis=-1
        )

    # Build RAG and apply normalized cut
    if rag is None:
        rag = rag_mean_feature(feats_map_rescaled, label, mode='similarity', sigma=rag_weight_sigma)
    labels = graph.cut_normalized(label, rag, thresh=ncut_thresh)

    # PCA for visualization
    h, w, c = feats_map_rescaled.shape
    flat_feats = feats_map_rescaled.reshape(-1, c)
    pca = PCA(n_components=3)
    if c <= 3:
        rgb_vis = feats_map_rescaled
    else:
        rgb_vis = pca.fit_transform(flat_feats).reshape(h, w, 3)
        rgb_vis = (rgb_vis - rgb_vis.min()) / (rgb_vis.max() - rgb_vis.min())

    grid_plot_list_imgs(
        images=[image,rgb_vis],
        ncols=2,
        fig_size=6
    )


    # Visualization
    fig, ax = plt.subplots(ncols=4, sharex=True, sharey=True, figsize=(24, 24))

    ax[0].set_title('Image')
    ax[0].imshow(image, cmap='gray' if image.ndim == 2 else None)

    ax[1].set_title('Superpixels')
    ax[1].imshow(mark_boundaries(rgb_vis, label,mode='inner'))

    ax[2].set_title('RAG')
    lc = graph.show_rag(
        label,
        rag,
        rgb_vis,
        border_color='yellow',
        img_cmap='gray',
        edge_cmap='coolwarm',
        edge_width=1,
        ax=ax[2]
    )

    ax[3].set_title('Normalized Cut')
    ax[3].imshow(color.label2rgb(labels, rgb_vis, kind='avg'))

    for a in ax:
        a.axis('off')

    plt.tight_layout()
    plt.show()



import math
import numpy as np
import matplotlib.pyplot as plt
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.neighbors import NearestNeighbors
from sklearn.cluster import spectral_clustering, KMeans
from sklearn.decomposition import PCA
from scipy.spatial.distance import pdist, squareform

from scipy.interpolate import splprep, splev
from skimage.measure import find_contours,label
from skimage.morphology import remove_small_objects,binary_erosion

"""
### get indexes list in a rectengle of shape (H,W)
indexes = [(i, j) for i in range(H) for j in range(W)]
print(indexes)
"""


def get_boundary_via_erosion(mask):
    # Step 1: Label the connected components if binary
    labeled_mask = label(mask)

    # Step 2: Extract boundaries of each region
    boundaries = np.zeros_like(labeled_mask, dtype=bool)

    for region_label in np.unique(labeled_mask):
        if region_label == 0:
            continue  # skip background

        # Create binary mask for this region
        region_mask = labeled_mask == region_label

        # Get the boundary using erosion
        eroded = binary_erosion(region_mask)
        boundary = region_mask ^ eroded  # XOR: region minus eroded = boundary

        # Store boundaries
        boundaries |= boundary  # add to global boundary mask
        
        return boundaries

def interplate_contour(contour, smoothing=0.01, num_points=400):
    x, y = contour[::100, 1], contour[::100, 0]
    tck, u = splprep([x, y], s=smoothing)
    u_new = np.linspace(0, 1, num_points)
    x_new, y_new = splev(u_new, tck)
    return np.vstack((y_new, x_new)).T

# Label connected components
def get_smooth_contours(mask,s=500,min_contour_point_num = 1000):

    """
    Extract and smooth contours from a mask using B-spline approximation.

    Parameters
    ----------
    mask : 2D np.ndarray
        input mask.
    s : float
        Smoothing factor for B-spline.
    min_contour_point_num : int, optional
        Minimum number of points required in a contour to be kept.

    Returns
    -------
    smoothed_contours : list of np.ndarray
        List of smoothed contour arrays of shape (N, 2).

    Example
    -------
    >>> smoothed_contours = get_smooth_contours(mask, s=5)
    >>> plt.imshow(mask, cmap='gray')
    >>> for sc in smoothed_contours:
    >>>     plt.plot(sc[:, 1], sc[:, 0], '-r', linewidth=0.5)
    >>> plt.axis('off')
    >>> plt.show()
    """

    labeled_mask = label(mask)

    # remove very small regions
    min_size = 50  # adjust this threshold
    labeled_mask = remove_small_objects(labeled_mask, min_size)

    # extract contour for each region
    contours = []
    for region_label in np.unique(labeled_mask):
        if region_label == 0:
            continue
        binary_region = (labeled_mask == region_label)
        cs = find_contours(binary_region, level=0.5)
        contours.extend(cs)  # Each region may have multiple contours

    #discard samll contour
    contours = [c for c in contours if c.shape[0]> min_contour_point_num]
    #smooth the contour using a b-spline to approximate
    smoothed_contours = [interplate_contour(c,smoothing=s) for c in contours]

    return smoothed_contours 




        
def three_pca_as_rgb_image(feats_list,final_image_shape):
    # PCA to 3 components
    """
    Applies Principal Component Analysis (PCA) to reduce a feature array to 3 components 
    and map the result into an RGB image for visualization.

    Parameters:
    ----------
    feats_list : np.ndarray
        A 2D array of shape (N, C), where N is the number of pixels (or spatial locations) 
        and C is the number of feature channels at each location. C must be >= 1.
    
    final_image_shape : tuple
        A tuple of two or three integers representing the shape (Z, Y, X) or (H, W, D) of the output 
        image volume. The product of these dimensions must match N.

    Returns:
    -------
    rgb_vis : np.ndarray
        A 4D array of shape (*final_image_shape, 3), where the last dimension represents
        the three principal components mapped to RGB channels. The values are normalized 
        to the range [0, 1] for display.
    """
    C = feats_list.shape[-1]

    if C <=3:
        return feats_list.reshape(*final_image_shape,C)
    
    pca = PCA(n_components=3)
    rgb_vis = pca.fit_transform(feats_list).reshape(*final_image_shape, 3)
    # Normalize for display
    rgb_vis = (rgb_vis - rgb_vis.min()) / (rgb_vis.max() - rgb_vis.min())

    return rgb_vis


def construct_knn_graph(feat_lst, knn, image_shape,spatial_neighbour_flag =False, distance_decay_flag=True,d_sigma=2):
    """
    """
    N = feat_lst.shape[0]
    loc_lst = list(np.ndindex(image_shape))

    similarity_matrix = cosine_similarity(feat_lst)
    np.fill_diagonal(similarity_matrix, 0)  # Remove self-loops

    # compute distance deacay
    dist = pdist(loc_lst, metric='euclidean')
    dist_matrix = squareform(dist)
    print(f"distacne_matrix: {dist_matrix.shape}")
    dist_matrix = np.exp(- dist_matrix**2/(2*d_sigma**2))

    if distance_decay_flag:
        similarity_matrix = similarity_matrix * dist_matrix
    else :
        similarity_matrix = similarity_matrix

    
    if spatial_neighbour_flag:
        vec_lst = loc_lst
        nn = NearestNeighbors(n_neighbors= knn +1, metric='euclidean') #self similarity has been removed
        nn.fit(vec_lst)
    else:
        vec_lst = feat_lst
        nn = NearestNeighbors(n_neighbors= knn +1, metric='cosine') #self similarity has been removed
        nn.fit(vec_lst)

    
    knn_graph = np.zeros_like(similarity_matrix)
    for i in range(N):
        neighbors = nn.kneighbors([vec_lst[i]], return_distance=False)[0]
        knn_graph[i, neighbors] = similarity_matrix[i, neighbors]
    
    return (knn_graph + knn_graph.T) / 2  # Ensure symmetry


def sc_plot_grid_results(encoded, img_size, knn_values, num_clusters_values,spatial_neighbour_flag = False ,spatial_decay_flag = True ,d_sigma=8):
    """
    Plot clustering results in a grid for different KNN and cluster numbers.

    Args:
        encoded (np.ndarray or torch.Tensor): Feature map of shape (N, C), where N = H * W.
        img_size (tuple): Size of the original image (H, W).
        knn_values (list of int): List of KNN values to try.
        num_clusters_values (list of int): List of number of clusters to try.

    Example:
        >>> num_clusters_values = [4, 6, 8]
        >>> knn_values = [4, 8, 16, 32]
        >>> sc_plot_grid_results(encoded, img_size, knn_values, num_clusters_values)
    """

    fig, axes = plt.subplots(len(knn_values), len(num_clusters_values), figsize=(len(num_clusters_values) * 4, len(knn_values) * 4))
    
    for i, knn in enumerate(knn_values):
        for j, num_clusters in enumerate(num_clusters_values):
            knn_graph = construct_knn_graph(feat_lst=encoded, knn=knn, image_shape=(img_size),spatial_neighbour_flag=spatial_neighbour_flag,distance_decay_flag=spatial_decay_flag,d_sigma=d_sigma)
            labels = spectral_clustering(knn_graph, n_clusters=num_clusters, eigen_solver="arpack")
            label_image = labels.reshape(img_size)
            
            ax = axes[i,j]
            ax.imshow(label_image, cmap='tab20')
            ax.set_title(f'sc: knn={knn}, clusters={num_clusters},spatial:{spatial_neighbour_flag},d_sigma:{d_sigma}')
            ax.axis('off')
    
    plt.tight_layout()
    plt.show()





def kmeans_grid_results(encoded,img_shape,K_values = [4, 6, 8, 12]):

    fig, axes = plt.subplots(1, len(K_values), figsize=(len(K_values) * 4, 4))  # One row, multiple columns

    for ax, K in zip(axes, K_values):
        # Apply K-means clustering
        kmeans = KMeans(n_clusters=K, random_state=42)
        labels = kmeans.fit_predict(encoded)
        
        # Display label image
        ax.imshow(labels.reshape(img_shape), cmap='tab20')
        ax.set_title(f'sklearn_{K} KMeans')
        ax.axis('off')

    plt.tight_layout()
    plt.show()

def grid_plot_list_imgs(images,col_labels=None,row_labels=None,ncols=3,fig_size=4,show=True):

    n_images = len(images)
    ncols = ncols
    nrows = math.ceil(n_images / ncols)

    fig, axes = plt.subplots(nrows, ncols, figsize=(fig_size*ncols,  fig_size*nrows))

    for i, ax in enumerate(axes.flat):
        if i < n_images:
            ax.imshow(images[i], cmap='gray')
        ax.axis('off')
            # Add column labels at top row

        if i < ncols:
            if col_labels:
                ax.set_title(col_labels[i], fontsize=14)

        # Add row labels at first column
        if i % ncols == 0:
            if row_labels:
                ax.set_ylabel(row_labels[i // ncols], fontsize=14, rotation=0, labelpad=40, va='center')


    if show:
        plt.tight_layout()
        plt.show()
    return fig